from importlib import import_module
import logging
from ..items import FieldManifest

logger = logging.getLogger(__name__)
extractor_classes = import_module(f'web_parsers.extractors.xml')


class XMLExtractorManifest:

    def __init__(self, extractor_type=None, extractor_id=None, extractor_fields=None):
        self.extractor_type = extractor_type
        try:
            extractor_cls = getattr(extractor_classes, extractor_type)
        except AttributeError as e:
            logger.error("Failed to import the extractor_type:{extractor_type} with error {error}".format(
                extractor_type=extractor_type,
                error=e
            ))
            extractor_cls = None
        self.extractor_cls = extractor_cls
        self.extractor_id = extractor_id
        if extractor_fields:
            self.extractor_fields = [FieldManifest(**item_manifest) for item_manifest in
                                     extractor_fields]
        else:
            self.extractor_fields = None

    def to_dict(self):
        return {
            "extractor_type": self.extractor_type,
            "extractor_id": self.extractor_id,
            "extractor_fields": self.extractor_fields,
        }

    def __repr__(self):
        return "<XMLExtractorManifest extractor_id=\"{extractor_id}\" " \
               "extractor_fields=\"{extractor_fields}\" >".format(extractor_id=self.extractor_id,
                                                                  extractor_fields=self.extractor_fields)



from web_parsers.extractors.base import ExtractorBase, ContentExtractorBase
from web_parsers.utils.selectors import extract_html_field
from web_parsers.utils.url import get_urn, get_domain, get_absolute_url
import json


class ParagraphsExtractor(ExtractorBase):
    def run(self):
        data = {}
        extracted_data = []
        elements = self.html_selector.css("p::text").extract()
        for el in elements:
            extracted_data.append(el)
        data[self.extractor_id] = [d.strip() for d in extracted_data]
        return data


class HeadingsExtractor(ExtractorBase):
    def run(self):
        data = {}
        extracted_data = []
        heading_tags = ["h1", "h2", "h3", "h4", "h5", "h6", "h7"]
        elements = self.html_selector.css("::text,".join(heading_tags)).extract()
        for el in elements:
            extracted_data.append(el)
        data[self.extractor_id] = [d.strip() for d in extracted_data if d.strip()]
        return data


class MainContentExtractor(ExtractorBase):
    # TODO - implement this
    pass


class TableContentExtractor(ExtractorBase):
    def run(self):
        data = {}
        tables = []
        for table in self.html_selector.css("table"):
            table_data = []
            table_headers = [th.extract() for th in table.css("thead tr th::text")]
            for row in table.css("tbody tr"):
                row_data = [td.extract() for td in row.css("td::text")]
                row_dict = dict(zip(table_headers, row_data))
                table_data.append(row_dict)
            tables.append(table_data)
        data[self.extractor_id] = tables
        return data


class MetaTagExtractor(ExtractorBase):
    def run(self):
        data = {}
        meta_data_dict = {}

        elements = self.html_selector.css('meta')
        for element in elements:
            # for open graph type of meta tags
            meta_property = element.xpath("@{0}".format('property')).extract_first()
            if meta_property:
                meta_property = meta_property.replace(":", "__").replace(".", "__")
                meta_data_dict[meta_property] = element.xpath(
                    "@{0}".format('content')).extract_first() or element.xpath("@{0}".format('value')).extract_first()

            meta_name = element.xpath("@{0}".format('name')).extract_first()
            if meta_name:
                meta_name = meta_name.replace(":", "__").replace(".", "__")
                meta_data_dict["meta__{}".format(meta_name)] = element.xpath(
                    "@{0}".format('content')).extract_first() or element.xpath("@{0}".format('value')).extract_first()

        try:
            title = self.html_selector.css('title::text').get()
            if title:
                meta_data_dict["title"] = title
        except Exception as e:
            pass
        data[self.extractor_id] = meta_data_dict
        return data


class CustomDataExtractor(ContentExtractorBase):
    extracted_data = {}

    def add_to_field(self, field_data=None, extractor_field=None):
        if extractor_field.flatten_data:
            if field_data is not None:
                self.extracted_data.update(field_data)
        else:
            self.extracted_data[extractor_field.field_id] = field_data

    def run(self):
        extractor_fields = self.extractor.extractor_fields
        for extractor_field in extractor_fields:
            nodes = self.html_selector.css(extractor_field.element_query.get("value"))
            child_selectors = extractor_field.child_selectors
            if extractor_field.data_type.startswith("List"):
                extracted_items = []
                if extractor_field.data_attribute == "element":
                    for node in nodes:
                        extracted_item = {}
                        for child_extractor in child_selectors:
                            extracted_item[child_extractor.field_id] = extract_html_field(
                                node, child_extractor
                            )
                        extracted_items.append(extracted_item)
                    self.add_to_field(field_data=extracted_items, extractor_field=extractor_field)
                else:
                    extracted_items = []
                    for node in nodes:
                        extracted_items.append(extract_html_field(
                            node, extractor_field
                        ))
                    self.add_to_field(field_data=extracted_items, extractor_field=extractor_field)

            else:
                if extractor_field.data_attribute == "element":
                    node = nodes[0]
                    extracted_item = {}
                    for child_extractor in child_selectors:
                        extracted_item[child_extractor.field_id] = extract_html_field(
                            node, child_extractor
                        )
                    self.add_to_field(field_data=extracted_item, extractor_field=extractor_field)

                else:
                    field_data = extract_html_field(
                        self.html_selector, extractor_field
                    )
                    self.add_to_field(field_data=field_data, extractor_field=extractor_field)

        return {self.extractor_id: self.extracted_data}

    # def run(self):
    #     data = {}
    #     extracted_data = {}
    #     for extractor_item in self.extractor.extractor_fields:
    #         if extractor_item.data_attribute == 'element' and extractor_item.child_selectors.__len__() > 0:
    #             # TODO - currently only support multiple elements strategy. what if multiple=False
    #             # review this; not sure if this is still applicable.
    #             elements = self.html_selector.css(extractor_item.element_query.get("value"))
    #             elements_data = []
    #             for el in elements:
    #                 datum = {}
    #                 for child_selector in extractor_item.child_selectors:
    #                     _d = extract_html_field(el, child_selector)
    #                     datum[child_selector.field_id] = _d if _d else None
    #                 elements_data.append(datum)
    #             data_type = extractor_item.data_type
    #             if data_type.startswith("List") is False:
    #                 single_data = elements_data[0]
    #                 extracted_data[extractor_item.field_id] = single_data
    #             else:
    #                 extracted_data[extractor_item.field_id] = elements_data
    #         else:
    #             _d = extract_html_field(self.html_selector, extractor_item)
    #             extracted_data[extractor_item.field_id] = _d
    #         if extractor_item.flatten_data is True:
    #             extracted_data[extractor_item.field_id] = self.flatten_data(extracted_data[extractor_item.field_id])
    #
    #     data[self.extractor_id] = extracted_data
    #     return data


class IconsExtractor(ExtractorBase):
    def run(self):
        data = {}
        meta_data_dict = {}

        favicon = self.html_selector.xpath('//link[@rel="shortcut icon"]').xpath("@href").get()
        if favicon:
            meta_data_dict['favicon'] = get_absolute_url(url=favicon, origin_url=self.url)

        elements = self.html_selector.xpath('//link[@rel="icon" or @rel="apple-touch-icon-precomposed"]')
        for element in elements:
            # for open graph type of meta tags
            meta_property = element.xpath("@sizes").extract_first()
            if meta_property:
                meta_property = meta_property.replace(":", "__").replace(".", "__")
                meta_data_dict[meta_property] = element.xpath("@{0}".format('href')).extract_first()
        data[self.extractor_id] = meta_data_dict
        return data


class JSONLDExtractor(ExtractorBase):

    def run(self):
        data = {}
        extracted_data = []

        elements = self.html_selector.xpath('//script[@type="application/ld+json"]/text()').extract()
        for element in elements:
            # for open graph type of meta tags
            try:
                element = element.strip()
                element = json.loads(element)
                extracted_data.append(element)
            except Exception as e:
                pass
        data[self.extractor_id] = extracted_data
        return data


class PlainHTMLContentExtractor(ExtractorBase):

    def run(self):
        return {self.extractor_id: self.html_selector.__str__()}


class FeedUrlExtractor(ExtractorBase):
    def run(self):
        data = {self.extractor_id: {}}
        data[self.extractor_id]['rss__xml'] = self.html_selector.xpath('//link[@type="application/rss+xml"]').xpath(
            "@href").extract_first()
        data[self.extractor_id]['rss__atom'] = self.html_selector.xpath('//link[@type="application/atom+xml"]').xpath(
            "@href").extract_first()
        return data


class PageOverviewExtractor(ContentExtractorBase):

    def run(self):
        data = {}

        meta_tags_data = MetaTagExtractor(
            url=self.url,
            html_selector=self.html_selector,
            extractor_id=self.extractor_id
        ).run().get(self.extractor_id, {})

        paragraphs_data = ParagraphsExtractor(
            url=self.url,
            html_selector=self.html_selector,
            extractor_id="paragraphs"
        ).run().get("paragraphs", {})
        # TODO - clean the data, that is extracted. ex:  meta_tags_data.get("title")
        extracted_data = {
            "title":
                meta_tags_data.get("title") or
                meta_tags_data.get("meta__title") or
                meta_tags_data.get("og__title") or
                meta_tags_data.get("fb__title") or
                meta_tags_data.get("meta__twitter__title"),
            "description":
                meta_tags_data.get("description") or
                meta_tags_data.get("meta__description") or
                meta_tags_data.get("og__description") or
                meta_tags_data.get("fb__description") or
                meta_tags_data.get("meta__twitter__description"),
            "image":
                meta_tags_data.get("image") or
                meta_tags_data.get("meta__image") or
                meta_tags_data.get("og__image") or
                meta_tags_data.get("fb__image") or
                meta_tags_data.get("meta__twitter__image"),
            "url":
                meta_tags_data.get("url") or
                meta_tags_data.get("meta__url") or
                meta_tags_data.get("og__url") or
                meta_tags_data.get("fb__url") or
                meta_tags_data.get("meta__twitter__url"),
            "page_type": meta_tags_data.get("og__type"),
            "keywords": meta_tags_data.get("meta__keywords"),
            "domain": get_domain(self.url),
            "first_paragraph": paragraphs_data[0] if len(paragraphs_data) > 0 else None,
            "shortlink_url": self.html_selector.xpath('//link[@rel="shortlink"]').xpath("@href").extract_first(),
            "canonical_url": self.html_selector.xpath('//link[@rel="canonical"]').xpath("@href").extract_first()
        }
        data[self.extractor_id] = extracted_data
        return data



from web_parsers.extractors.base import ExtractorBase, ContentExtractorBase
import logging

logger = logging.getLogger(__name__)


class PythonExtractorManifest(ContentExtractorBase):
    """
Usage :


extractor_type: PythonExtractorManifest
extractor_id: python_extractor
extractor_fn: 'def extractor_fn(html_selector=None): return {"html_selector": html_selector.__str__()}'


    """

    def run(self):
        extractor_fn = self.extractor.get("extractor_fn")
        data = {}
        global_fns = {}
        exec(extractor_fn.strip(), global_fns)
        simulate_fn = global_fns['extractor_fn']
        try:
            data[self.extractor_id] = simulate_fn(html_selector=self.html_selector)
        except Exception as e:
            logger.error("Failed to run the PythonExtractorManifest  with error {}".format(e))
            data[self.extractor_id] = None
        return data




from .single import StringField, IntField, FloatField, DictField, RawField
import logging

logger = logging.getLogger(__name__)


class ListFieldTransformerBase(object):
    transformer_class = None

    def __init__(self, data=None):
        self.data = data
        self.validate_transformer()

    def validate_transformer(self):
        if self.transformer_class is None:
            raise NotImplementedError("transformer_class should be assigned for List type of FieldTransformer")

    def try_or_none(self):

        result_data_list = []
        for d in self.data:
            try:
                result_data = self.transformer_class(data=d).transform()
                result_data_list.append(result_data)
            except Exception as e:
                logger.debug("Failed to transform data `{d}` to `{klass_name}` type with "
                             "error `{error}`".format(d=d, klass_name=self.__class__.__name__, error=e))
        return result_data_list

    def transform(self):
        return self.try_or_none()


class ListStringField(ListFieldTransformerBase):
    transformer_class = StringField


class ListIntField(ListFieldTransformerBase):
    transformer_class = IntField


class ListFloatField(ListFieldTransformerBase):
    transformer_class = FloatField


class ListDictField(ListFieldTransformerBase):
    transformer_class = DictField


class ListRawField(ListFieldTransformerBase):
    transformer_class = RawField



def validate_extractor(extractor):
    validation_errors = []
    if extractor.get("extractor_id") is None:
        validation_errors.append("extraction_id cannot be None in extractor {extractor}".format(extractor=extractor))
    if extractor.get("extractor_type") is None:
        validation_errors.append("extractor_type cannot be None in extractor {extractor}".format(extractor=extractor))

    return validation_errors



from .images import ImagesExtractor
from .links import AllLinksExtractor, AllLinksAnalyticsExtractor
from .content import CustomDataExtractor, PageOverviewExtractor, \
    ParagraphsExtractor, HeadingsExtractor, TableContentExtractor, MetaTagExtractor, IconsExtractor, JSONLDExtractor, \
    FeedUrlExtractor, PlainHTMLContentExtractor, MainContentExtractor



import logging

logger = logging.getLogger(__name__)


class ExtractorBase:
    """

    """

    def __init__(self, url=None, html_selector=None, extractor_id=None, **kwargs):
        """
        :param html_selector: html html_selector of the request
        :param extractor_id: the field with which data is stored in database.
        """
        # if url is None:
        #     logger.warning(
        #         "url cannot be blank, as it is used to generate absolute urls in most of the cases, you can "
        #         "give http://dummy-url.com as url, if you want to give a dummy url"
        #     )
        if None in [html_selector, extractor_id]:
            raise Exception(
                "Invalid input to the extractor class, html_selector, extractor and extractor_id are mandatory")

        self.url = url
        self.html_selector = html_selector
        self.extractor_id = extractor_id

    def run(self):
        raise NotImplementedError()


class ContentExtractorBase(ExtractorBase):
    """

    """

    def __init__(self, url=None, html_selector=None, extractor_id=None, extractor=None):
        super(ContentExtractorBase, self).__init__(url=url, html_selector=html_selector, extractor_id=extractor_id)
        self.extractor = extractor



import logging
from web_parsers.utils import convert_html_to_selector
from .base import ParserBase

logger = logging.getLogger(__name__)


class HTMLParser(ParserBase):
    """

    extractor_manifest should be a list of extractors ie., `web_parsers.extractors.
    """
    selector_key = "html_selector"

    def __init__(self, url=None, string_data=None, extractor_manifest=None):
        super().__init__(url=url, string_data=string_data, extractor_manifest=extractor_manifest)
        if self.url is None:
            raise Exception(
                "url cannot be blank, as it is used to generate absolute urls in most of the cases, you can "
                "give http://dummy-url.com as url, if you want to give a dummy url"
            )

    def parse_data(self, string_data):
        """
        this function will be used to convert string to html tree.
        :return:
        """
        return convert_html_to_selector(string_data)



import importlib
from parsel import Selector


def convert_html_to_selector(html):
    return Selector(html)


class SelectorExtractor(object):

    @staticmethod
    def get_list_data(elements=None):
        data_cleaned = []
        data = elements.extract()
        for i, datum in enumerate(data):
            if datum:
                data_cleaned.append(datum.strip())
        return data_cleaned

    @staticmethod
    def get_single_data(elements=None):
        data = elements.extract_first()
        if data:
            return data.strip()
        return data


def transform_data(data=None, data_type=None):
    fields = importlib.import_module("web_parsers.fields")
    Klass = getattr(fields, data_type)
    data = Klass(data=data).transform()
    return data


def clean_data(elements=None, item_extractor=None):
    """

    This is where are the extracted data will be cleaned up and applied functions and data types as needed.

    :param elements:
    :param item_extractor:
    :return:
    """

    # TODO - list is calculated
    data_type = item_extractor.data_type
    if data_type.startswith("List"):
        multiple = True
    else:
        multiple = False

    data_extractor = SelectorExtractor()
    if multiple is True:
        extracted_data = data_extractor.get_list_data(elements=elements)
    else:
        extracted_data = data_extractor.get_single_data(elements=elements)
    data = transform_data(data=extracted_data, data_type=data_type)
    return data


def extract_html_field(html_selector, item_extractor):
    element_query = item_extractor.element_query
    if item_extractor.data_attribute in ['text']:
        if element_query.get("type") == 'css':
            elements = html_selector.css("{0}::{1}".format(element_query.get('value'),
                                                          item_extractor.data_attribute))
            return clean_data(elements=elements, item_extractor=item_extractor)
        else:
            elements = html_selector.xpath("{0}/{1}".format(element_query.get('value'),
                                                           item_extractor.data_attribute))
            return clean_data(elements=elements, item_extractor=item_extractor)
    elif item_extractor.data_attribute == 'html':
        if element_query.get('type') == 'css':
            elements = html_selector.css(element_query.get('value'))
            return clean_data(elements=elements, item_extractor=item_extractor)
        else:
            elements = html_selector.xpath("{0}/{1}".format(element_query.get('value'),
                                                           item_extractor.data_attribute))
            return clean_data(elements=elements, item_extractor=item_extractor)
    else:
        if element_query.get('type') == 'css':
            elements = html_selector.css(element_query.get('value')) \
                .xpath("@{0}".format(item_extractor.data_attribute))
            return clean_data(elements=elements, item_extractor=item_extractor)
        else:
            elements = html_selector.xpath("{0}/{1}".format(element_query.get('value'),
                                                           item_extractor.data_attribute))
            return clean_data(elements=elements, item_extractor=item_extractor)



import xmltodict as xd
import json
from web_parsers.utils.xml import get_nodes, parse_xml, get_node_text, get_node_html
import logging
from .base import ParserBase

logger = logging.getLogger(__name__)


class XMLParser(ParserBase):
    """
from web_parsers.parsers.xml import XMLParser
from web_parsers.utils.other import yaml_to_json, generate_random_id
import pprint
import urllib.request

string_data = urllib.request.urlopen("https://invana.io/feed.xml").read()

xml_extractor_yml = \"""
- extractor_type: CustomDataExtractor
  extractor_id: channel_info
  extractor_fields:
  - field_id: channel
    element_query:
      type: xpath
      value: '/rss/channel'
    data_attribute: element
    data_type: DictField
    child_selectors:
      - field_id: title
        element_query:
          type: xpath
          value: title
        data_attribute: text
        data_type: StringField
      - field_id: description
        element_query:
          type: xpath
          value: description
        data_attribute: text
        data_type: StringField
  - field_id: pages
    element_query:
      type: xpath
      value: '/rss/channel/item'
    data_attribute: element
    data_type: ListDictField
    child_selectors:
      - field_id: title
        element_query:
          type: xpath
          value: title
        data_attribute: text
        data_type: StringField
      - field_id: description
        element_query:
          type: xpath
          value: description
        data_attribute: text
        data_type: StringField
      - field_id: link
        element_query:
          type: xpath
          value: link
        data_attribute: text
        data_type: StringField
      - field_id: guid
        element_query:
          type: xpath
          value: guid
        data_attribute: text
        data_type: StringField
      - field_id: is_perma_link
        element_query:
          type: xpath
          value: guid
        data_attribute: isPermaLink
        data_type: StringField
\"""
xml_extractor_manifest = yaml_to_json(xml_extractor_yml)

xml_parser = XMLParser(string_data=string_data, extractor_manifest=xml_extractor_manifest)

# usage 1
result = xml_parser.run_extractors(flatten_extractors=True)

# usage 2
result = to_dict()


    """
    selector_key = "html_selector"

    def to_dict(self):
        return json.loads(json.dumps(xd.parse(self.string_data)))

    def parse_data(self, string_data):
        """
        this function will be used to convert html/xml tree.
        :return:
        """
        return parse_xml(string_data)



from web_parsers.extractors.base import ExtractorBase, ContentExtractorBase
from web_parsers.utils.url import get_urn, get_domain


class AllLinksExtractor(ExtractorBase):
    def run(self):
        data = {}
        extracted_data = []
        links = self.html_selector.css("a").xpath("@href").extract()

        for link in links:
            if link and not link.startswith("#"):
                extracted_data.append(link)
        data[self.extractor_id] = extracted_data
        return data


class AllLinksAnalyticsExtractor(ContentExtractorBase):

    def run(self):
        data = {}
        extracted_data = AllLinksExtractor(
            url=self.url,
            html_selector=self.html_selector,
            extractor=self.extractor,
            extractor_id="all_links"
        ).run().get("all_links", {})

        links_data = {}
        for link in extracted_data:
            domain = get_domain(link)
            if domain in links_data.keys():
                links_data[domain].append(link)
            else:
                links_data[domain] = [link]

        data[self.extractor_id] = [{"domain": domain, "links": domain_links, "links_count": domain_links.__len__()} for
                                   domain, domain_links in links_data.items()]

        return data


class SameDomainLinkExtractor(ExtractorBase):
    pass


class ForeignDomainLinkExtractor(ExtractorBase):
    pass


class CustomLinkExtractor(ExtractorBase):
    pass



import logging

logger = logging.getLogger(__name__)


class ParserBase:
    selector_key = None

    def __init__(self, string_data, url=None, extractor_manifest=None):
        self.string_data = string_data
        self.url = url
        self.extractor_manifest = extractor_manifest



    def parse_data(self, string_data):
        """
        this function will be used to convert string to html/xml tree.
        :return:
        """
        raise NotImplementedError()

    def get_selector_key(self):
        if self.selector_key is None:
            raise Exception("selector_key should be assigned to Parser classes")
        return self.selector_key

    def run_extractor(self, xml_tree=None, extractor=None):
        extractor_id = extractor.extractor_id
        logger.info("Running extractor:'{}' on url:{}".format(extractor_id, self.url))
        try:
            extractor_object = extractor.extractor_cls(
                **{
                    'url': self.url,
                    self.get_selector_key(): xml_tree,
                    'extractor': extractor,
                    'extractor_id': extractor_id
                }
            )
            return extractor_object.run()
        except Exception as error:
            logger.error(
                "Failed to extract data from  the extractor '{extractor_id}:{extractor_type}' on url "
                "'{url}' with error: '{error}'".format(
                    extractor_id=extractor_id,
                    extractor_type=extractor.extractor_type,
                    url=self.url,
                    error=error)
            )
            return {extractor_id: None}

    @staticmethod
    def flatten_extracted_data(all_extracted_data):
        all_extracted_data_new = {}
        for k, v in all_extracted_data.items():
            all_extracted_data_new.update(v)
        return all_extracted_data_new

    def run_extractors(self, flatten_extractors=False):
        xml_tree = self.parse_data(self.string_data)
        all_extracted_data = {}
        for extractor in self.extractor_manifest.extractors:
            extracted_data = self.run_extractor(extractor=extractor, xml_tree=xml_tree)
            all_extracted_data.update(extracted_data)
        if flatten_extractors is True:
            return self.flatten_extracted_data(all_extracted_data)
        return all_extracted_data



from lxml import etree


def parse_xml(string_data):
    return etree.fromstring(string_data)


def get_nodes(xml_tree, xpath):
    return xml_tree.xpath(xpath)


def get_node_text(node):
    return node.text or '' + ''.join(
        etree.tostring(e) for e in node)


def get_node_html(node):
    return etree.tostring(node, pretty_print=False)


def extract_xml_field(node, extractor_config):
    child_node = node.xpath(extractor_config.element_query.get("value").lstrip("/"))[0]
    if extractor_config.data_attribute == "text":
        return get_node_text(child_node)
    elif extractor_config.data_attribute == "html":
        return get_node_html(child_node)
    else:
        return child_node.attrib.get(
            extractor_config.data_attribute
        )



import logging
from .alpha.html import HTMLExtractorManifest
from .alpha.xml import XMLExtractorManifest

logger = logging.getLogger(__name__)


class Owner:

    def __init__(self, title=None, ownership_type="Individual", email=None, website_url=None):
        self.title = title
        self.Individual = ownership_type
        self.email = email
        self.website_url = website_url


class WebParserManifest:
    valid_versions = ("alpha",)
    valid_parser_types = ("html", "xml")
    """
    Usage


    WebParserManifest(
        title="invana.io blogs",
        domain="invana.io",
        version="alpha",
        test_urls="https://invana.io/blogs",
        extractors=[],
        owner={
            "title": "Ravi Raja Merugu",
            "ownership_type": "Individual",
            "email": "rrmerugu@gmail.com",
            "website_url": "https://rrmerugu.github.io"
        }
    )
    """

    def __init__(self,
                 title=None,
                 parser_type="html",
                 domain=None,
                 version="alpha",
                 test_urls=None,
                 extractors=None,
                 owner=None,
                 ):

        if version not in self.valid_versions:
            raise Exception("version can only be '{}' for now".format(",".join(self.valid_versions)))
        if parser_type not in self.valid_parser_types:
            raise Exception("parser_type can only be '{}' for now".format(",".join(self.valid_parser_types)))

        self.title = title
        self.test_urls = test_urls
        self.version = version
        self.parser_type = parser_type
        self.domain = domain
        if owner:
            self.owner = Owner(**owner)
        else:
            self.owner = owner

        manifest_cls = None
        if self.parser_type == "html":
            manifest_cls = HTMLExtractorManifest
        elif self.parser_type == "xml":
            manifest_cls = XMLExtractorManifest

        self.extractors = []

        for extractor in extractors:
            self.extractors.append(manifest_cls(**extractor))

    def to_dict(self):
        raise NotImplementedError()



class ExtractedItem:

    def __init__(self, _id=None, field_name=None, data=None):
        self.data = data if data else None
        self._id = _id
        self.field_name = field_name

    def get_dict(self):
        """
        returns the data in json format
        :return:
        """
        if self.field_name:
            return {
                self.field_name: self.data
            }
        else:
            return self.data

    def __repr__(self):
        return "<ExtractedItem field_name=\"{field_name}\" data=\"{data}\" >".format(
            field_name=self.field_name,
            data=self.data
        )



import locale
import re
import logging
import json
import ast

logger = logging.getLogger(__name__)


class FieldTransformerBase(object):
    """
    IntField(data="1").transform()

    """

    def get_method(self):
        """
        Should return the function that takes data as input. Ex: str, int, float,

        :return:
        """
        raise NotImplementedError("This method is not implemented")

    def __init__(self, data=None):
        self.data = data

    def try_or_none(self):
        transformation_method = self.get_method()
        if transformation_method:
            try:
                result_data = transformation_method(self.data)
            except Exception as e:
                logger.debug(e)
                result_data = self.data
            return result_data
        else:
            return self.data

    def transform(self):
        return self.try_or_none()


class StringField(FieldTransformerBase):

    def get_method(self):
        return str


class IntField(FieldTransformerBase):
    def get_method(self):
        def custom_int(data):
            if type(data) is str:
                data = locale.atoi(data)
            return int(data)
        return custom_int


class FloatField(FieldTransformerBase):
    def get_method(self):
        def custom_float(data):
            data = re.findall(r"[-+]?\d*\.\d+|[-+]?\d+", data)
            if len(data) > 0:
                return float(data[0])
            else:
                return float(0)

        return custom_float


class DictField(FieldTransformerBase):

    @staticmethod
    def dict_or_json_dump(d):
        if type(d) == dict:
            return d
        elif type(d) == str:
            return json.loads(d)
        return d

    def get_method(self):
        return  ast.literal_eval


class RawField(FieldTransformerBase):
    def get_method(self):
        return None



#!/usr/bin/env python

from setuptools import setup, find_packages

setup(
    name='web-parsers',
    version='0.0.2',
    description='Simple, extendable HTML and XML data extraction engine using YAML configurations '
                'and some times pythonic functions.',
    author='Ravi Raja Merugu',
    author_email='ravi@invanalabs.ai',
    url='https://github.com/invanalabs/web_parsers',
    packages=find_packages(
        exclude=("dist", "docs", "tests",)
    ),
    install_requires=[
        'parsel>=1.5.2',
        'PyYAML>=5.3.1',
        'xmltodict==0.12.0'
    ],
    entry_points={
    },
    python_requires='>=3.6',
    classifiers=[
        'Development Status :: 3 - Alpha',
        # Chose either "3 - Alpha", "4 - Beta" or "5 - Production/Stable" as the current state of your package
        'Environment :: Console',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Build Tools',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: POSIX :: Linux',
        'Operating System :: MacOS',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: Implementation :: CPython',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Software Development :: Libraries :: Application Frameworks',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
)



from ..base import ContentExtractorBase
from web_parsers.utils.xml import extract_xml_field, get_nodes


class CustomDataExtractor(ContentExtractorBase):
    extracted_data = {}

    def add_to_field(self, field_data=None, extractor_field=None):
        if extractor_field.flatten_data:
            if field_data is not None:
                self.extracted_data.update(field_data)
        else:
            self.extracted_data[extractor_field.field_id] = field_data

    def run(self):
        extractor_fields = self.extractor.extractor_fields
        for extractor_field in extractor_fields:
            nodes = get_nodes(
                self.html_selector,
                extractor_field.element_query.get("value")
            )
            child_selectors = extractor_field.child_selectors
            if extractor_field.data_type.startswith("List"):
                extracted_items = []
                if extractor_field.data_attribute == "element":
                    for node in nodes:
                        extracted_item = {}
                        for child_extractor in child_selectors:
                            extracted_item[child_extractor.field_id] = extract_xml_field(
                                node, child_extractor
                            )
                        extracted_items.append(extracted_item)
                    self.add_to_field(field_data=extracted_items, extractor_field=extractor_field)
                else:
                    extracted_items = []
                    for node in nodes:
                        extracted_items.append(extract_xml_field(
                            node, extractor_field
                        ))
                    self.add_to_field(field_data=extracted_items, extractor_field=extractor_field)

            else:
                if extractor_field.data_attribute == "element":
                    node = nodes[0]
                    extracted_item = {}
                    for child_extractor in child_selectors:
                        extracted_item[child_extractor.field_id] = extract_xml_field(
                            node, child_extractor
                        )
                    self.add_to_field(field_data=extracted_item, extractor_field=extractor_field)

                else:
                    field_data = extract_xml_field(
                        self.html_selector, extractor_field
                    )
                    self.add_to_field(field_data=field_data, extractor_field=extractor_field)

        return {self.extractor_id: self.extracted_data}



class FieldManifest:

    def __init__(self,
                 field_id=None,
                 data_type="RawField",
                 element_query=None,
                 data_attribute=None,
                 child_selectors=None,
                 flatten_data=False
                 ):
        self.field_id = field_id
        self.data_type = data_type
        self.element_query = element_query
        self.data_attribute = data_attribute

        if (data_attribute != "element" and flatten_data is True) or (
                data_attribute == "element" and data_type != "DictField" and flatten_data is True):
            raise Exception(
                "error on `{field_id}` field manifest : flatten_data=True, can only be used with "
                "data_attribute=='element' and data_type=='DictField'".format(field_id=field_id)
            )

        self.flatten_data = flatten_data

        if child_selectors:
            self.child_selectors = [FieldManifest(**child_selector) for child_selector in
                                    child_selectors]
        else:
            self.child_selectors = []

    def __repr__(self):
        return "<FieldManifest data_type='{data_type}' " \
               "element_query='{element_query}' " \
               "data_attribute='{data_attribute}' " \
               ">".format(data_type=self.data_type, element_query=self.element_query,
                          data_attribute=self.data_attribute)

    def to_dict(self):
        return {
            "field_id": self.field_id,
            "element_query": self.element_query,
            "data_type": self.data_type,
            "data_attribute": self.data_attribute
        }
